`timescale 1ns / 100ps

/*
 * verilog model of 65C02 CPU with 6502MMU.
 *
 * Based on original 6502 "Arlet 6502 Core" by Arlet Ottens 
 * 
 * (C) Arlet Ottens, <arlet@c-scape.nl>
 *
 * Feel free to use this code in any project (commercial or not), as long as you
 * keep this message, and the copyright notice. This code is provided "as is", 
 * without any warranties of any kind.
 * 
 * Support for 65C02 instructions and addressing modes by David Banks and Ed Spittles
 *
 * (C) 2016 David Banks and Ed Spittles
 *
 * Feel free to use this code in any project (commercial or not), as long as you
 * keep this message, and the copyright notice. This code is provided "as is", 
 * without any warranties of any kind.
 *
 * Support for 6502MMU instructions, bank handling and WDC/Rockwell instructions by Renato Bugge
 *
 * (C) 2022 Renato Bugge
 *
 * Feel free to use this code in any project (commercial or not), as long as you
 * keep this message, and the copyright notice. This code is provided "as is", 
 * without any warranties of any kind.
 *  
 */

/*
 * Note that not all 6502 interface signals are supported (yet).  The goal
 * is to create an Acorn Atom model, and the Atom didn't use all signals on
 * the main board.
 *
 * The data bus is implemented as separate read/write buses. Combine them
 * on the output pads if external memory is required.
 */

`define SIM

/*
 * Two things were needed to correctly implement 65C02 NOPs
 * 1. Ensure the microcode state machine uses an appropriate addressing mode for the opcode length
 * 2. Ensure there are no side-effects (e.g. register updates, memory stores, etc)
 * 
 * If IMPLEMENT_NOPS is defined, the state machine is modified accordingly.
 */

`define IMPLEMENT_NOPS

/*
 * MMU implementation uses NOP instructions in 6502/65C02, but they are still NOPS unless the MMU
 * is switched on. Compared to the 6502MMU with a real MOS6502, R6502, 65C02, it should handle
 * the same way with the following instructions:
 *
 * $82 $xy	    MMS $xy (MMS=Memory Management System register):
 
 * $xy=%10101010 to switch MMU to STANDBY. Any other value for x switches MMU OFF.
 * $xy=%01010101 to switch MMU ON if it is in STANDBY or ON.
 * when MMU is ON, $xy must be %0zzzzzzz on the recurring accesses, or MMU will be switched OFF:
 
 * $xy=%01zz zzzz to switch on MMJ_STORAGE, which means that JBANK is used for all writes
 * $xy=%00zz zzzz to switch off MMJ_STORAGE, which means that FBANK is used for all writes
 * $xy=%0z1z zzzz to switch on BANK_REG_MEM, which shows registers in the $0000-$01ff location
 * $xy=%0z0z zzzz to switch off BANK_REG_MEM, which hides the registers from memory
 * $xy=%0zz1 zzzz to switch on IGNORE_LSB, which ignores the 4 LSB of the BANK register
 * $xy=%0zz0 zzzz to switch off IGNORE_LSB, which means that all bits in the BANK register are used
 * $xy=%0zzz 1zzz to switch interrupt bank instruction to NBANK mode
 * $xy=%0zzz 0zzz to switch interrupt bank instruction to IBANK mode
 * $xy=%0zzz z1zz to swap $44/$54 instructions to MMJ $ZP and MMF $ZP,X (SWAP=1)
 * $xy=%0zzz z0zz to keep $44/$54 instructions to MMF $ZP and MMJ $ZP,X (SWAP=0)
 * $xy=%0zzz zz1z to switch on FULL_BLOCK_MOVE, which means that BANK is used to move full 64KB address range
 * $xy=%0zzz zz0z to switch off FULL_BLOCK_MOVE, which means that BLOCK register sets 4KB address range for MMU
 * $xy=%0zzz zzz1 to switch on PROTECTED_MEMORY, which means that memory is accessed according to PTABLE (see below)
 * $xy=%0zzz zzz0 to switch off PROTECTED_MEMORY, which means that all memory is accessible ("supervisor mode")

 * When BANK_REG_MEM=ON, the followring registers can be manipulated directly at memory:
 *   $0000 = FBANK (e.g. the bank we will fetch to/from in the next instruction(s))
 *   $0001 = current ZBANK (e.g. ZBANK(JBANK) which is were the current Zeropage resides)
 *   $0002 = ZBANK for previous JBANK (e.g. ZBANK location after RTS or RTI)
 *   $0003 = IBANK register
 *   $0004 = NBANK register
 *   $0005 = JBPOINTER (e.g. JBANK stack pointer) as a 8-bit value
 *   $0100-$01FF = full JBANK stack as 255 different 8-bit pointers
 * (Future : $0006-$01FF = full JBANK stack as 253 different 16-bit pointers)
 * The location of these registers is not affected by JBANK, ZBANK or FBANK
 * These registers can only be written to in Supervisor mode.
 * In User mode, the MMS instruction will always be %011zzzzz or %010zzzzz, e.g. only bit 5 can be changed.

 * $C2 $xx      MMZ $xx   (MMZ=Memory Management ZBANK register)
 
 * $xx = LSB of BANK at which the ZeroPage ($0000-$00FF) will reside if ZEROPAGE_BANK=ON
 * IGNORE_LSB does not affect the MMZ instruction. ZeroPage will move as either a 4KiB ($xx0FF) or 64KiB ($xx00FF) block
 * Also affects Stack so that Stack is moved to block $xx100-$xx1FF (or $xx0100-$xx01FF in 64KiB block mode)
 * The MMZ instruction will use the FBANK (see below) as index for storage into the ZBANK register array. Thus ZBANK[FBANK] is
 * modified. There are two ways the ZBANK register is used according to PROTECTED_MEMORY setting:
 * 1) PROTECTED_MEMORY=ON. In Supervisor Mode, ZBANK[0] is used as the ZBANK register, while ZBANK[JBANK]
 *    is the active register in User mode.
 * 2) PROTECTED_MEMORY=OFF. ZBANK[JBANK] is always used as the ZBANK register.
 * Remember to save stack pointer before using this command (or just clear it).
  
 * $E2 #$xy		MMB #$xy (MMB=Memory Management BLOCK setup)
 * BLOCK register $xy were y is a value 0-15 which sets memory location $y000-$yfff as MMU memory if FULL_BLOCK_MOVE=OFF
 * Any memory outside of this range is used as ordinary 6502 memory without being affected by the MMU
 * 
 * If PROTECTED_MEMORY=ON this instruction is used to set up PTABLE (see below)
 
 * $FC $xx $yy	MMF $yyxx (MMF=Memory Management Fetch BANK)
 * The instructions that follow will read or store from Memory Bank FBANK=$yyxx
 
 * $44 $ZP		MMF $ZP (MMF=Memory Management Fetch BANK)
 * Reads ZP (Zero Page) location and puts value into FBANK
 * The instruction that follow will read or store from Memory Bank FBANK=$00xx
 * ZBANK register will affect were the Zero Page location is and can therefore affect this instruction.
 * IGNORE_LSB does not affect the MMF $ZP instruction.

 * $DC $xx $yy 	MMJ $yyxx (MMJ=Memory Management Jump BANK)
 * The instruction that follow will jump (JMP or JSR) to Memory Bank JBANK=$yyxx
 * Note that current memory bank is stored in the Jump Stack, so the next RTS will return to the current
 * memory bank.
 * PS: Jump stack not implemented yet, so RTS returns to whatever address is in stack+JBANK

 * $54 $ZP		MMJ $ZP,X (MMJ=Memory Management Jump BANK)
 * Reads ZP (Zero Page) location and puts value into JBANK
 * The instruction that follow will jump (JMP or JSR) to Memory Bank JBANK=$00xx
 * IGNORE_LSB does not affect the MMJ $ZP instruction.

 * $5C $xx $yy 	MMI $yyxx (MMI=Memory Management Interrupt BANK)
 * Sets IBANK (IRQ Bank register) or NBANK (NMI Bank register) 
 * During the next IRQ/NMI interrupt, the memory location will be set to Bank $xxyy
 * Note that if the memory location is outside MMU memory (for FULL_BLOCK_MOVE=OFF), the IRQ will
 * not be affected.
 
 * ACCESS KEY setup:
 * To set up, turn PROTECTED_MEMORY=ON (MMU goes into supervisor mode)
 * $E2 #$xx: (MMB #$xx)
 
 * Sets ACCESS key to $xx for FBANK (can only be set in Supervisor mode)
 * If key not set to zero and in user mode, code running in a bank must have the same key as another
 * bank for it to read/write or execute data there.	If this is not the case, the read/write/jmp
 * will be redirected back to the current JBANK.
 * Usage:
 *	- A MMF is used before the MMB to set the key
 *	- Setting the keys can only be done in "supervisor mode" ( e.g. ACCESS_KEY=0 and PROTECTED_MEMORY is on).
 *  - After setup is complete, "user mode" can be started by a JSR to a BANK with ACCESS_KEY>0
 *	- A protected bank can only JMP to BANK if the ACCESS key of the two BANKS are the same.
 *  - To exit a protected BANK, an RTS back to an unprotected BANK (with execute key zero) will resume "supervisor mode".
 *	  During an interrupt, the cpu may come from a protected bank. 
 
 *	  Current implementation is limited to 20 bits (1MiB) which is 16 BANKs of 64KiB or 256 BANKs of 4KiB.
 *    For 64KiB BANKs, only key numbers 0 to 15 are used (20 bits address)
 
 *	  Future: 24-bit implementation with 256 banks of 64KiB or 65535 BANKs of 4KiB (requires larger BANK numbers).


 * Other notes
 * MMJ changes mmu_block[block_pnt+1]
 * block_pnt increases on JSR $Axxx
 * JMP changes current mmu_block: mmu_block[block_pnt]=mmu_block[block_pnt+1] (no change in block_pnt)
 * block_pnt decreases on RTS
 * MMI will set the Interrupt vector Upper Bits (IUB)

What is implemented now:  
Partially:   $82 $xy	    MMS $xy (MMS=Memory Management System register):
Implemented: $E2 #$xy		MMB #$xy  (MMB=Memory Management BLOCK register or PTABLE setup)
Implemented: $C2 $xx        MMZ $xx   (MMZ=Memory Management ZBANK register)
Implemented: $FC $xx $yy	MMF $yyxx (MMF=Memory Management Fetch BANK)
Implemented: $DC $xx $yy 	MMJ $yyxx (MMJ=Memory Management Jump BANK)
Implemented: $44 $ZP		MMF $ZP (MMF=Memory Management Fetch BANK)  
Implemented: $54 $ZP,X		MMJ $ZP (MMF=Memory Management Jump BANK)
Not implem.: $F4
Not implem.: $D4
Not implem.: $5C (MMI)

BANK_REG_MEM is not implemented fully (but control by MMS is)
BANK_REG_MEM for reading registers implementet but not tested
BANK_REG_MEM for writing registers implementet but not tested 

Rockwell/WDC instructions (not fully tested):
Implemented: $x7 (x=$0-$7) 	RMB0-7 $ZP
Implemented: $x7 (x=$8-$F) 	SMB0-7 $ZP
Implemented: $xF (x=$0-$7) 	BBR0-7 $ZP,$BR
Implemented: $xF (x=$8-$F) 	BBS0-7 $ZP,$BR
Implemented: $CB			WAI
Implemented: $DB			STP


 */

`define IMPLEMENT_MMU

/*
 * Two things were needed to correctly implement 65C02 BCD arithmentic
 * 1. The Z flag needs calculating over the BCD adjusted ALU output
 * 2. The N flag needs calculating over the BCD adjusted ALU output
 *
 * If IMPLEMENT_CORRECT_BCD_FLAGS is defined, this additional logic is added
 */

// `define IMPLEMENT_CORRECT_BCD_FLAGS

module cpu_65c02( clk, reset, AB, DI, DO, WE, IRQ, NMI, RDY, WAI);

input clk;              // CPU clock 
input reset;            // reset signal
output reg [19:0] AB;   // address bus (20-bit, future: 24-bit or 32-bit)
input [7:0] DI;         // data in, read bus
output [7:0] DO;        // data out, write bus
output WE;              // write enable
input IRQ;              // interrupt request
input NMI;              // non-maskable interrupt request
input RDY;              // Ready signal. Pauses CPU when RDY=0
output reg WAI;				// WAI output. Needs to pull RDY low (from external code) to stop CPU.
/*
 * internal signals
 */

reg  [19:0] PC;         // Program Counter (as wide as address but)
reg  [7:0] ABM;         // Address Bus Register MMU
reg  [7:0] ABL;         // Address Bus Register LSB
reg  [7:0] ABH;         // Address Bus Register MSB
wire [7:0] ADD;         // Adder Hold Register (registered in ALU)

reg  [7:0] DIHOLD;      // Hold for Data In
reg  DIHOLD_valid;      //
wire [7:0] DIMUX;       //

reg  [7:0] IRHOLD;      // Hold for Instruction register 
reg  IRHOLD_valid;      // Valid instruction in IRHOLD

reg  [7:0] AXYS[3:0];    // A, X, Y and S register file

reg  [7:0] FBANK;		// Fetch Memory bank
reg  [7:0] IBANK;		// IRQ memory bank
reg  [7:0] NBANK;		// NMI memory bank
reg	 [7:0] ZBANK [0:255]; // ZP&stack memory bank ARRAY
reg  [7:0] JBANK [0:255]; // JSR memory bank stack
reg  [7:0] TBANK;		// Temporary bank register
reg	 [7:0] JBPOINTER;	// Which JBANK value to save
reg  [1:0] MMU_STATE;	//=0 for off, =1 for standby =3 for ON.
reg  SWAP;				//=1 to swap $44 and $54 instructions (MMJ/MMF)
reg  MMJ_STORAGE;		//=1 if writing goes to MMJ block instead of MMF block
reg  BANK_REG_MEM;		//=1 to show different BANK registers in ZP/stack
reg  BANK_REG_MEM_WE;   //=1 if writing to BANK registers in ZP/stack
reg  [9:0] BANK_REG_ADDR;  // Temporary address
reg  [7:0] BANK_REG_DATA;  // Temporary data
reg  BANK_REG_ACT;  	// Flag to change register
reg  IGNORE_LSB;		//=0 to ingnore bit 0-3 of the FBANK register (=uses bit 4->)
reg  FULL_BLOCK_MOVE;	//=1 if full 64KB block is moved with BANK, =0 if 4KB block is moved with BANK
reg  PROTECTED_MEMORY;	//=1 if using PTABLE, =0 if not used
reg  NIBANK;			//=0 for IBANK (IRQ) , =1 for NBANK (NMI)
reg  [3:0] BANK_RANGE;	//=0..15 is MSB for 4KB memory block that moves with BANK (unused for 64KB block)
reg  [7:0] ACCESS_KEY [0:255]; // READ/WRITE/EXECUTE key for protected memory (one key per BANK)
//reg  [7:0] WRITE_KEY [0:255]; // WRITE keys for protected memory
//reg  [7:0] EXECUTE_KEY [0:255]; // EXECUTE keys for protected memory


reg  C = 0;             // carry flag (init at zero to avoid X's in ALU sim)
reg  Z = 0;             // zero flag
reg  I = 0;             // interrupt flag
reg  D = 0;             // decimal flag
reg  V = 0;             // overflow flag
reg  N = 0;             // negative flag
wire AZ;                // ALU Zero flag
wire AZ1;               // ALU Zero flag (BCD adjusted)
reg  AZ2;               // ALU Second Zero flag, set using TSB/TRB semantics
wire AV;                // ALU overflow flag
wire AN;                // ALU negative flag
wire AN1;               // ALU negative flag (BCD adjusted)
wire HC;                // ALU half carry

reg  [7:0] AI;          // ALU Input A
reg  [7:0] BI;          // ALU Input B
wire [7:0] DI;          // Data In
wire [7:0] IR;          // Instruction register
reg  [7:0] DO;          // Data Out 
wire [7:0] AO;          // ALU output after BCD adjustment
reg  WE;                // Write Enable
reg  CI;                // Carry In
wire CO;                // Carry Out 
wire [7:0] PCM = {4'b0000,PC[19:16]};
wire [7:0] PCH = PC[15:8];
wire [7:0] PCL = PC[7:0];

reg NMI_edge = 0;       // captured NMI edge

reg [1:0] regsel;                       // Select A, X, Y or S register
wire [7:0] regfile = AXYS[regsel];      // Selected register output

parameter 
        SEL_A    = 2'd0,
        SEL_S    = 2'd1,
        SEL_X    = 2'd2, 
        SEL_Y    = 2'd3;

/*
 * define some signals for watching in simulator output
 */


`ifdef SIM
wire [7:0]   A = AXYS[SEL_A];           // Accumulator
wire [7:0]   X = AXYS[SEL_X];           // X register
wire [7:0]   Y = AXYS[SEL_Y];           // Y register 
wire [7:0]   S = AXYS[SEL_S];           // Stack pointer 
`endif

wire [7:0] P = { N, V, 2'b11, D, I, Z, C };

/*
 * instruction decoder/sequencer
 */

reg [5:0] state;

/*
 * control signals
 */

reg PC_inc;             // Increment PC
reg [19:0] PC_temp;     // intermediate value of PC 

reg [1:0] src_reg;      // source register index
reg [1:0] dst_reg;      // destination register index

reg index_y;            // if set, then Y is index reg rather than X 
reg load_reg;           // loading a register (A, X, Y, S) in this instruction
reg inc;                // increment
reg write_back;         // set if memory is read/modified/written 
reg load_only;          // LDA/LDX/LDY instruction
reg store;              // doing store (STA/STX/STY)
reg adc_sbc;            // doing ADC/SBC
reg compare;            // doing CMP/CPY/CPX
reg shift;              // doing shift/rotate instruction
reg rotate;             // doing rotate (no shift)
reg backwards;          // backwards branch
reg cond_true;          // branch condition is true
reg [3:0] cond_code;    // condition code bits from instruction
reg [4:0] cond_save;    // condition code bits for BBS/BBR
reg shift_right;        // Instruction ALU shift/rotate right 
reg alu_shift_right;    // Current cycle shift right enable
reg [3:0] op;           // Main ALU operation for instruction
reg [3:0] alu_op;       // Current cycle ALU operation 
reg adc_bcd;            // ALU should do BCD style carry 
reg adj_bcd;            // results should be BCD adjusted
reg [7:0] clear_bit;	// RMB instruction
reg	[7:0] set_bit;		// SMB instruction
reg [7:0] bc_bit;		// BBR instruction
reg	[7:0] bs_bit;		// BBS instruction


/* 
 * some flip flops to remember we're doing special instructions. These
 * get loaded at the DECODE state, and used later
 */
reg store_zero;         // doing STZ instruction
reg trb_ins;            // doing TRB instruction
reg txb_ins;            // doing TSB/TRB instruction
reg bit_ins;            // doing BIT instruction
reg bit_ins_nv;         // doing BIT instruction that will update the n and v flags (i.e. not BIT imm)
reg plp;                // doing PLP instruction
reg php;                // doing PHP instruction 
reg clc;                // clear carry
reg sec;                // set carry
reg cld;                // clear decimal
reg sed;                // set decimal
reg cli;                // clear interrupt
reg sei;                // set interrupt
reg clv;                // clear overflow 
reg brk;                // doing BRK

reg [1:0] mmu_fetch;	// set bank flags for next fetch instruction(s)
reg [1:0] mmu_jump;		// set bank flags for next jump instruction
reg [2:0] mmu_special;	// MMS/MMB/MMI instruction flags

reg res;                // in reset

/*
 * ALU operations
 */

parameter
        OP_OR  = 4'b1100,
        OP_AND = 4'b1101,
        OP_EOR = 4'b1110,
        OP_ADD = 4'b0011,
        OP_SUB = 4'b0111,
        OP_ROL = 4'b1011,
        OP_A   = 4'b1111;

/*
 * Microcode state machine. Basically, every addressing mode has its own
 * path through the state machine. Additional information, such as the
 * operation, source and destination registers are decoded in parallel, and
 * kept in separate flops. 
 */

parameter 
    ABS0   = 7'd0,  // ABS     - fetch LSB      
    ABS1   = 7'd1,  // ABS     - fetch MSB
    ABSX0  = 7'd2,  // ABS, X  - fetch LSB and send to ALU (+X)
    ABSX1  = 7'd3,  // ABS, X  - fetch MSB and send to ALU (+Carry)
    ABSX2  = 7'd4,  // ABS, X  - Wait for ALU (only if needed)
    BRA0   = 7'd5,  // Branch  - fetch offset and send to ALU (+PC[7:0])
    BRA1   = 7'd6,  // Branch  - fetch opcode, and send PC[15:8] to ALU 
    BRA2   = 7'd7,  // Branch  - fetch opcode (if page boundary crossed)
    BRK0   = 7'd8,  // BRK/IRQ - push PCH, send S to ALU (-1)
    BRK1   = 7'd9,  // BRK/IRQ - push PCL, send S to ALU (-1)
    BRK2   = 7'd10, // BRK/IRQ - push P, send S to ALU (-1)
    BRK3   = 7'd11, // BRK/IRQ - write S, and fetch @ fffe
    DECODE = 7'd12, // IR is valid, decode instruction, and write prev reg
    FETCH  = 7'd13, // fetch next opcode, and perform prev ALU op
    INDX0  = 7'd14, // (ZP,X)  - fetch ZP address, and send to ALU (+X)
    INDX1  = 7'd15, // (ZP,X)  - fetch LSB at ZP+X, calculate ZP+X+1
    INDX2  = 7'd16, // (ZP,X)  - fetch MSB at ZP+X+1
    INDX3  = 7'd17, // (ZP,X)  - fetch data 
    INDY0  = 7'd18, // (ZP),Y  - fetch ZP address, and send ZP to ALU (+1)
    INDY1  = 7'd19, // (ZP),Y  - fetch at ZP+1, and send LSB to ALU (+Y) 
    INDY2  = 7'd20, // (ZP),Y  - fetch data, and send MSB to ALU (+Carry)
    INDY3  = 7'd21, // (ZP),Y) - fetch data (if page boundary crossed)
    JMP0   = 7'd22, // JMP     - fetch PCL and hold
    JMP1   = 7'd23, // JMP     - fetch PCH
    JMPI0  = 7'd24, // JMP IND - fetch LSB and send to ALU for delay (+0)
    JMPI1  = 7'd25, // JMP IND - fetch MSB, proceed with JMP0 state
    JSR0   = 7'd26, // JSR     - push PCH, save LSB, send S to ALU (-1)
    JSR1   = 7'd27, // JSR     - push PCL, send S to ALU (-1)
    JSR2   = 7'd28, // JSR     - write S
    JSR3   = 7'd29, // JSR     - fetch MSB
    PULL0  = 7'd30, // PLP/PLA/PLX/PLY - save next op in IRHOLD, send S to ALU (+1)
    PULL1  = 7'd31, // PLP/PLA/PLX/PLY - fetch data from stack, write S
    PULL2  = 7'd32, // PLP/PLA/PLX/PLY - prefetch op, but don't increment PC
    PUSH0  = 7'd33, // PHP/PHA/PHX/PHY - send A to ALU (+0)
    PUSH1  = 7'd34, // PHP/PHA/PHX/PHY - write A/P, send S to ALU (-1)
    READ   = 7'd35, // Read memory for read/modify/write (INC, DEC, shift, RMB/SMB)
    REG    = 7'd36, // Read register for reg-reg transfers
    RTI0   = 7'd37, // RTI     - send S to ALU (+1)
    RTI1   = 7'd38, // RTI     - read P from stack 
    RTI2   = 7'd39, // RTI     - read PCL from stack
    RTI3   = 7'd40, // RTI     - read PCH from stack
    RTI4   = 7'd41, // RTI     - read PCH from stack
    RTS0   = 7'd42, // RTS     - send S to ALU (+1)
    RTS1   = 7'd43, // RTS     - read PCL from stack 
    RTS2   = 7'd44, // RTS     - write PCL to ALU, read PCH 
    RTS3   = 7'd45, // RTS     - load PC and increment
    WRITE  = 7'd46, // Write memory for read/modify/write 
    ZP0    = 7'd47, // Z-page  - fetch ZP address
    ZPX0   = 7'd48, // ZP, X   - fetch ZP, and send to ALU (+X)
    ZPX1   = 7'd49, // ZP, X   - load from memory
    IND0   = 7'd50, // (ZP)    - fetch ZP address, and send to ALU (+0)
    JMPIX0 = 7'd51, // JMP (,X)- fetch LSB and send to ALU (+X)
    JMPIX1 = 7'd52, // JMP (,X)- fetch MSB and send to ALU (+Carry)
    JMPIX2 = 7'd53, // JMP (,X)- Wait for ALU (only if needed)
	MMJ0   = 7'd54, // MMJ     - Store ZP in JBANK, increase JBPOINTER
	MMJ1   = 7'd55, // MMJ     - Store LSB/MSB in JBANK, increase JBPOINTER
	MMF0   = 7'd56, // MMF     - Store ZP in FBANK
	MMF1   = 7'd57, // MMF     - Store LSB/MSB in FBANK
	ZB0    = 7'd58, // Z-page  - fetch ZP address before branch
	ZB1    = 7'd59, // Z-page  - fetch ZP address before branch
	ZB2    = 7'd60, // Z-page  - fetch ZP address before branch
	ZB3    = 7'd61, // Z-page  - fetch ZP address before branch
	ZB4    = 7'd62, // Z-page  - fetch ZP address before branch
	WAI0   = 7'd63, // WAI     - Wait for an interrupt or reset
	WAI1   = 7'd64, // WAI     - Wait for an interrupt or reset
	WAI2   = 7'd65, // WAI     - Wait for an interrupt or reset
	STP    = 7'd66; // STP     - Stop until reset

`ifdef SIM

/*
 * easy to read names in simulator output
 */
reg [8*6-1:0] statename;

always @*
    case( state )
            DECODE: statename = "DECODE";
            REG:    statename = "REG";
            ZP0:    statename = "ZP0";
            ZPX0:   statename = "ZPX0";
            ZPX1:   statename = "ZPX1";
            ABS0:   statename = "ABS0";
            ABS1:   statename = "ABS1";
            ABSX0:  statename = "ABSX0";
            ABSX1:  statename = "ABSX1";
            ABSX2:  statename = "ABSX2";
            IND0:   statename = "IND0";
            INDX0:  statename = "INDX0";
            INDX1:  statename = "INDX1";
            INDX2:  statename = "INDX2";
            INDX3:  statename = "INDX3";
            INDY0:  statename = "INDY0";
            INDY1:  statename = "INDY1";
            INDY2:  statename = "INDY2";
            INDY3:  statename = "INDY3";
             READ:  statename = "READ";
            WRITE:  statename = "WRITE";
            FETCH:  statename = "FETCH";
            PUSH0:  statename = "PUSH0";
            PUSH1:  statename = "PUSH1";
            PULL0:  statename = "PULL0";
            PULL1:  statename = "PULL1";
            PULL2:  statename = "PULL2";
            JSR0:   statename = "JSR0";
            JSR1:   statename = "JSR1";
            JSR2:   statename = "JSR2";
            JSR3:   statename = "JSR3";
            RTI0:   statename = "RTI0";
            RTI1:   statename = "RTI1";
            RTI2:   statename = "RTI2";
            RTI3:   statename = "RTI3";
            RTI4:   statename = "RTI4";
            RTS0:   statename = "RTS0";
            RTS1:   statename = "RTS1";
            RTS2:   statename = "RTS2";
            RTS3:   statename = "RTS3";
            BRK0:   statename = "BRK0";
            BRK1:   statename = "BRK1";
            BRK2:   statename = "BRK2";
            BRK3:   statename = "BRK3";
            BRA0:   statename = "BRA0";
            BRA1:   statename = "BRA1";
            BRA2:   statename = "BRA2";
            JMP0:   statename = "JMP0";
            JMP1:   statename = "JMP1";
            JMPI0:  statename = "JMPI0";
            JMPI1:  statename = "JMPI1";
            JMPIX0: statename = "JMPIX0";
            JMPIX1: statename = "JMPIX1";
            JMPIX2: statename = "JMPIX2";
			MMJ0: 	statename = "MMJ0";
			MMJ1: 	statename = "MMJ1";
			MMF0: 	statename = "MMF0";
			MMF1: 	statename = "MMF1";
            ZB0:    statename = "ZB0";
            ZB1:    statename = "ZB1";
			ZB2:    statename = "ZB2";
			ZB3:    statename = "ZB3";
			ZB4:    statename = "ZB4";
            WAI0:    statename = "WAI0";
            WAI1:    statename = "WAI1";
            WAI2:    statename = "WAI2";
            STP:    statename = "STP";
			
    endcase

//always @( PC )
//      $display( "%t, PC:%04x IR:%02x A:%02x X:%02x Y:%02x S:%02x C:%d Z:%d V:%d N:%d P:%02x", $time, PC, IR, A, X, Y, S, C, Z, V, N, P );

`endif


/*
 * Program Counter Increment/Load. First calculate the base value in
 * PC_temp.
 */
always @*
    case( state )
        DECODE:         if( ((~I & IRQ) | NMI_edge) & ~(state==STP) )
                            PC_temp = {  FULL_BLOCK_MOVE?{ABM[3:0],ABH}:
										(BANK_RANGE==ABH[7:4])?{ABM[7:0],ABH[3:0]}:{4'd0,ABH}, ABL };
                        else
                            PC_temp = PC;


        JMPI1,
        JMPIX1,
        JMP1:           PC_temp = { FULL_BLOCK_MOVE?{JBANK[JBPOINTER][3:0],DIMUX}:
						(BANK_RANGE==DIMUX[7:4])?{JBANK[JBPOINTER][7:0],DIMUX[3:0]}:{4'd0,DIMUX}, ADD };
        JSR3:           PC_temp = { FULL_BLOCK_MOVE?{JBANK[JBPOINTER][3:0],DIMUX}:
						(BANK_RANGE==DIMUX[7:4])?{JBANK[JBPOINTER][7:0],DIMUX[3:0]}:{4'd0,DIMUX}, ADD };
        RTS3,           
        RTI4:           PC_temp = { FULL_BLOCK_MOVE?{JBANK[JBPOINTER-8'd1][3:0],DIMUX}:
						(BANK_RANGE==DIMUX[7:4])?{JBANK[JBPOINTER-8'd1][7:0],DIMUX[3:0]}:{4'd0,DIMUX}, ADD };
        ZB3,                
        BRA1:           PC_temp = { FULL_BLOCK_MOVE?{JBANK[JBPOINTER][3:0],ABH[7:0]}:
						(BANK_RANGE==ABH[7:4])?{JBANK[JBPOINTER][7:0],ABH[3:0]}:{ABM[3:0],ABH}, ADD };

        JMPIX2,
        BRA2:           PC_temp = {  FULL_BLOCK_MOVE?{JBANK[JBPOINTER][3:0],ADD}:
						(BANK_RANGE==ADD[7:4])?{JBANK[JBPOINTER][7:0],ADD[3:0]}:{4'd0,ADD}, PCL };

        BRK2:           PC_temp =      res ? 20'h0fffc : 
                                  NMI_edge ? (FULL_BLOCK_MOVE?{NBANK[3:0],16'hfffa}:{NBANK[7:0],12'hffa}): 
								  (FULL_BLOCK_MOVE?{IBANK[3:0],16'hfffe}:{IBANK[7:0],12'hffe});
		WAI2,
		STP:			PC_temp = PC;

        default:        PC_temp = PC;
    endcase

/*
 * Determine wether we need PC_temp, or PC_temp + 1
 */
always @*
    case( state )
        DECODE:         if( (~I & IRQ) | NMI_edge )
                            PC_inc = 0;
                        else
							begin
								PC_inc = 1;
								WAI = 1'b0;
							end

        ABS0,
        JMPIX0,
        JMPIX2,
        ABSX0,
        FETCH,
        BRA0,
        BRA2,
        BRK3,
        JMPI1,
        JMP1,
        RTI4,
		MMF0,
		MMF1,
		MMJ0,
		MMJ1,
        RTS3,
		WAI2:           PC_inc = 1;

        JMPIX1:         PC_inc = ~CO;       // Don't increment PC if we are going to go through JMPIX2
		ZB3,
        BRA1:           PC_inc = CO ^~ backwards;
		
		STP:			PC_inc = 0;			// Do nothing
		WAI0:			WAI = 1'b1;			// Pull WAI high

        default:        PC_inc = 0;
    endcase

/* 
 * Set new PC
 */
always @(posedge clk)
//	if( reset )
//        PC <= 16'hfffc;
    if( RDY )
        PC <= PC_temp + PC_inc;

/*
 * Address Generator 
 */

parameter
        ZEROPAGE  = 8'h00,
        STACKPAGE = 8'h01;

/*
 * The MMU constrains here are only for access FROM a MMU block towards non-MMU memory
 * Constrains between blocks are handled by MMF and MMJ	(which won't change block in user mode)
 */		
always @*
    case( state )
        INDX3,
        INDY2,
        JMPIX1,
        JMP1,
        JMPI1:			AB = { FULL_BLOCK_MOVE?{ABM[3:0],DIMUX}:     // 64KiB block?
						(BANK_RANGE==DIMUX[7:4])?{ABM[7:0],DIMUX[3:0]}:	   // within 4KiB block?
						(PROTECTED_MEMORY==1'b1 & ACCESS_KEY[JBANK[JBPOINTER]] > 8'd0)?    //normal mem & protected mem?
						{JBANK[JBPOINTER],DIMUX}:{4'b0000,DIMUX},ADD};	   // access <64KiB mem or prevent bank change
        RTI4:           AB = { FULL_BLOCK_MOVE?{ABM[3:0],DIMUX}:     // 64KiB block?
						(BANK_RANGE==DIMUX[7:4])?{ABM[7:0],DIMUX[3:0]}:	   // within 4KiB block?
						{4'b0000,DIMUX},ADD};		// RTI always from supervisor memory
        ABSX1,
        ABS1:           AB = { FULL_BLOCK_MOVE?{FBANK[3:0],DIMUX}:
						(BANK_RANGE==DIMUX[7:4])?{FBANK[7:0],DIMUX[3:0]}:
						(PROTECTED_MEMORY==1'b1 & ACCESS_KEY[JBANK[JBPOINTER]] > 8'd0)?
						{JBANK[JBPOINTER],DIMUX}:{4'b0000,DIMUX},ADD};

        BRA2,
        JMPIX2:          AB = { FULL_BLOCK_MOVE?{ABM[3:0],ADD}:
						(BANK_RANGE==ADD[7:4])?{ABM[7:0],ADD[3:0]}:
						(PROTECTED_MEMORY==1'b1 & ACCESS_KEY[JBANK[JBPOINTER]] > 8'd0)?
						{JBANK[JBPOINTER],ADD}:{4'b0000,ADD},ABL};
        INDY3,
        ABSX2:          AB = { FULL_BLOCK_MOVE?{FBANK[3:0],ADD}:
						(BANK_RANGE==ADD[7:4])?{FBANK[7:0],ADD[3:0]}:
						(PROTECTED_MEMORY==1'b1 & ACCESS_KEY[JBANK[JBPOINTER]] > 8'd0)?
						{JBANK[JBPOINTER],ADD}:{4'b0000,ADD},ABL};
        ZB3,
		BRA1:           AB = { ABM, ABH, ADD };		//Stay within bank

        JSR0,
        PUSH1,
        RTS0,
        RTI0,
        BRK0:           AB = { FULL_BLOCK_MOVE?{ZBANK[JBANK[JBPOINTER][7:0]][3:0],STACKPAGE}
							:{ZBANK[JBANK[JBPOINTER][7:0]][7:0],STACKPAGE[3:0]}, regfile };

        BRK1,
        JSR1,
        PULL1,
        RTS1,
        RTS2,
        RTI1,
        RTI2,
        RTI3,
        BRK2:           AB = { FULL_BLOCK_MOVE?{ZBANK[JBANK[JBPOINTER][7:0]][3:0],STACKPAGE}
							:{ZBANK[JBANK[JBPOINTER][7:0]][7:0],STACKPAGE[3:0]}, ADD };	   // Zeropage or stack access
        
        INDY1,
        INDX1,
        ZPX1,
        INDX2:          AB = { FULL_BLOCK_MOVE?{ZBANK[JBANK[JBPOINTER][7:0]][3:0],ZEROPAGE}
							:{ZBANK[JBANK[JBPOINTER][7:0]][7:0],ZEROPAGE[3:0]}, ADD };

        ZP0,
		ZB0,
        INDY0:          AB = { FULL_BLOCK_MOVE?{ZBANK[JBANK[JBPOINTER][7:0]][3:0],ZEROPAGE}
							:{ZBANK[JBANK[JBPOINTER][7:0]][7:0],ZEROPAGE[3:0]}, DIMUX };

        REG:        	AB = {ABM, ABH, ABL };		// Only register changes, keep within bank
		
        READ:          AB = { FULL_BLOCK_MOVE?{ABM[3:0],ABH}:
						(BANK_RANGE==ABH[7:4])?{ABM[7:0],ABH[3:0]}:
						(PROTECTED_MEMORY==1'b1 & ACCESS_KEY[JBANK[JBPOINTER]] > 8'd0)?
						{JBANK[JBPOINTER],ABH}:{4'b0000,ABH},ABL};
        WRITE:          AB = { FULL_BLOCK_MOVE?{(MMJ_STORAGE?JBANK[JBPOINTER][3:0]:ABM[3:0]),ABH}:
						(BANK_RANGE==ABH[7:4])?{(MMJ_STORAGE?JBANK[JBPOINTER][7:0]:ABM[7:0]),ABH[3:0]}:
						(PROTECTED_MEMORY==1'b1 & ACCESS_KEY[JBANK[JBPOINTER]] > 8'd0)?
						{JBANK[JBPOINTER],ABH}:{4'b0000,ABH},ABL};

        default:        AB = PC;
    endcase

/*
 * ABH/ABL pair is used for registering previous address bus state.
 * This can be used to keep the current address, freeing up the original
 * source of the address, such as the ALU or DI.
 */
always @(posedge clk)
    if( state != PUSH0 && state != PUSH1 && RDY && 
        state != PULL0 && state != PULL1 && state != PULL2 )
    begin
        ABL <= AB[7:0];
        ABH <= AB[15:8];
		ABM <= AB[19:16];	  	// This is the BANK part of the address
    end

/*
 * Data Out MUX 
 */
always @*
	begin
		case( state )
		WRITE:   DO = ADD;

		JSR0,
		BRK0:    DO = PCH;

		JSR1,
		BRK1:    DO = PCL;

		PUSH1:   DO = php ? P : ADD;

		BRK2:    DO = (IRQ | NMI_edge) ? (P & 8'b1110_1111) : P;

		default: DO = store_zero ? 0 : regfile;
		endcase
		if (state == WRITE & BANK_REG_MEM_WE == 1'b1)
			begin
				BANK_REG_DATA <= ADD;
				BANK_REG_ADDR <= AB[9:0];
				BANK_REG_ACT <= 1'b1;
			end					
		else BANK_REG_ACT <= 1'b0;
	end

/*
 * Write Enable Generator
 */

always @*
    case( state )
        BRK0,   // writing to stack or memory
        BRK1,
        BRK2,
        JSR0,
        JSR1,
        PUSH1:   WE = 1;
				// Special case if BANK_REG_MEM=ON and we access $0000-$01FF:
        WRITE:	begin
					BANK_REG_MEM_WE = ((BANK_REG_MEM == 1'b1) & (AB[15:8] < 8'd2));
					WE = BANK_REG_MEM_WE ? 0: 1 ;
				end
        INDX3,  // only if doing a STA, STX or STY
        INDY3,
        ABSX2,
        ABS1,
        ZPX1,
        ZP0:    begin
					BANK_REG_MEM_WE = ((BANK_REG_MEM == 1'b1) & (AB[15:8] < 8'd2) & store);
					WE = BANK_REG_MEM_WE ? 0: store ;
				end

        default: begin
					BANK_REG_MEM_WE = 0;
					WE = 0;
				end
    endcase

/*
 * register file, contains A, X, Y and S (stack pointer) registers. At each
 * cycle only 1 of those registers needs to be accessed, so they combined
 * in a small memory, saving resources.
 */

reg write_register;             // set when register file is written

always @*
    case( state )
        DECODE: write_register = load_reg & ~plp;

        PULL1, 
         RTS2, 
         RTI3,
         BRK3,
         JSR0,
         JSR2 : write_register = 1;

       default: write_register = 0;
    endcase

/*
 * BCD adjust logic
 */

always @(posedge clk)
    adj_bcd <= adc_sbc & D;     // '1' when doing a BCD instruction

reg [3:0] ADJL;
reg [3:0] ADJH;

// adjustment term to be added to ADD[3:0] based on the following
// adj_bcd: '1' if doing ADC/SBC with D=1
// adc_bcd: '1' if doing ADC with D=1
// HC     : half carry bit from ALU
always @* begin
    casex( {adj_bcd, adc_bcd, HC} )
         3'b0xx: ADJL = 4'd0;   // no BCD instruction
         3'b100: ADJL = 4'd10;  // SBC, and digital borrow
         3'b101: ADJL = 4'd0;   // SBC, but no borrow
         3'b110: ADJL = 4'd0;   // ADC, but no carry
         3'b111: ADJL = 4'd6;   // ADC, and decimal/digital carry
    endcase
end

// adjustment term to be added to ADD[7:4] based on the following
// adj_bcd: '1' if doing ADC/SBC with D=1
// adc_bcd: '1' if doing ADC with D=1
// CO     : carry out bit from ALU
always @* begin
    casex( {adj_bcd, adc_bcd, CO} )
         3'b0xx: ADJH = 4'd0;   // no BCD instruction
         3'b100: ADJH = 4'd10;  // SBC, and digital borrow
         3'b101: ADJH = 4'd0;   // SBC, but no borrow
         3'b110: ADJH = 4'd0;   // ADC, but no carry
         3'b111: ADJH = 4'd6;   // ADC, and decimal/digital carry
    endcase
end

assign AO = { ADD[7:4] + ADJH, ADD[3:0] + ADJL };

`ifdef IMPLEMENT_CORRECT_BCD_FLAGS

assign AN1 = AO[7];
assign AZ1 = ~|AO;

`else

assign AN1 = AN;
assign AZ1 = AZ;

`endif
   
/*
 * write to a register. Usually this is the (BCD corrected) output of the
 * ALU, but in case of the JSR0 we use the S register to temporarily store
 * the PCL. This is possible, because the S register itself is stored in
 * the ALU during those cycles.
 */
 
/* 
 * All MMU commands will get here, so handle accordingly.
 */
always @(posedge clk)
    if( write_register & RDY )
        AXYS[regsel] <= (state == JSR0) ? DIMUX : AO;

always @(negedge clk)
	if( reset == 1'b1 )
	begin
        FBANK <= 8'd0;
		JBANK[8'd0] <= 8'd0;
		JBANK[8'd1] <= 8'd0;
		JBPOINTER <= 8'd0;
		ZBANK[8'd0] <= 8'd0;
		IBANK <= 8'd0;
		NBANK <= 8'd0;
		TBANK <= 8'd0;
		MMU_STATE <= 2'b00;
		IGNORE_LSB <= 1'b0;
		NIBANK <= 1'b0;
		SWAP <= 1'b0;
		FULL_BLOCK_MOVE <= 1'b1;
		PROTECTED_MEMORY <= 1'b0;
		BANK_REG_MEM <= 1'b0;
		MMJ_STORAGE <= 1'b0;
//		AXYS[SEL_S] <=8'hff;
	end
	else if ( (state == FETCH) & RDY )
	begin
		if ( (mmu_fetch == 2'd1) & SWAP)
			JBANK[JBPOINTER+8'd1] <= {DIMUX}; 	// For 20 bit no preceding zeros, insert if needed (>20 bit)
		else if ( (mmu_jump == 2'd1) & SWAP
			& (ACCESS_KEY[JBANK[JBPOINTER]] == 8'd0 
			| ACCESS_KEY[JBANK[JBPOINTER]] == ACCESS_KEY[DIMUX]
			| PROTECTED_MEMORY == 1'b0))  // Check if protected
			FBANK <= {DIMUX};				// For 20 bit no preceding zeros, insert if needed (>20 bit)
		else if ( (mmu_fetch == 2'd1)
			& (ACCESS_KEY[JBANK[JBPOINTER]] == 8'd0 
			| ACCESS_KEY[JBANK[JBPOINTER]] == ACCESS_KEY[DIMUX]
			| PROTECTED_MEMORY == 1'b0))  // Check if protected
			FBANK <= {DIMUX};				// For 20 bit no preceding zeros, insert if needed (>20 bit)
		else if ( (mmu_jump == 2'd1) )
			JBANK[JBPOINTER+8'd1] <= {DIMUX}; 	// For 20 bit no preceding zeros, insert if needed (>20 bit)
		else if ( (mmu_special==3'b001) & ((ACCESS_KEY[JBANK[JBPOINTER]] == 8'd0)
			| (PROTECTED_MEMORY == 1'b0)))
			   casex ( {DIMUX,MMU_STATE} )						// MMS instruction
					10'b1010_1010_xx:   MMU_STATE <= 2'b01;		// MMU STANDBY
					10'b0101_0101_01:   MMU_STATE <= 2'b11;		// MMU ON
					10'b0xxx_xxxx_11:   begin
											MMJ_STORAGE <= DIMUX[6:6];
											BANK_REG_MEM <= DIMUX[5:5];
											IGNORE_LSB <= DIMUX[4:4];
											NIBANK <= DIMUX[3:3];
											SWAP <= DIMUX[2:2];
											FULL_BLOCK_MOVE <= DIMUX[1:1];
											PROTECTED_MEMORY <= DIMUX[0:0];
										end
						default:	MMU_STATE <= 2'b00;			// MMU OFF
				endcase
		else if ( (mmu_special==3'b010) )	//MMZ
			ZBANK[FBANK] <= DIMUX;
		else if ( (mmu_special==3'b100) & (NIBANK==1'b0) )	//MMI
			IBANK <= DIMUX;
		else if ( (mmu_special==3'b100) & (NIBANK==1'b1) )	//MMI
			NBANK <= DIMUX;
		else if ( (mmu_special==3'b011) & (MMU_STATE == 2'b11) & (PROTECTED_MEMORY == 1'b0)) //MMB
			BANK_RANGE <= DIMUX[3:0];
		else if ( (mmu_special==3'b011) & (MMU_STATE == 2'b11) & (PROTECTED_MEMORY == 1'b1)
			& (ACCESS_KEY[JBANK[JBPOINTER]] == 8'd0)) //Access key setup from supervisor
			ACCESS_KEY[FBANK] <= DIMUX;
	end
	else if ( (state == ABS1) & RDY & (MMU_STATE == 2'b11))
	begin
		if ( (mmu_fetch == 2'd2) 
			& (ACCESS_KEY[JBANK[JBPOINTER]] == 8'd0 
			| ACCESS_KEY[JBANK[JBPOINTER]] == ACCESS_KEY[{IGNORE_LSB?AB[11:4]:AB[7:0]}]
			| PROTECTED_MEMORY == 1'b0))  // Check if protected
			FBANK <= {IGNORE_LSB?AB[11:4]:AB[7:0]};	 // Comes from $ADDR and not the data in ADDR
		else if ( (mmu_jump == 2'd2) )
			JBANK[JBPOINTER+8'd1] <= {IGNORE_LSB?AB[11:4]:AB[7:0]}; 	// Comes from $ADDR and not the data in ADDR
	end
	else if ( (state == JMP1) & RDY & (MMU_STATE == 2'b11)
		& (ACCESS_KEY[JBANK[JBPOINTER]] == 8'd0 
		| ACCESS_KEY[JBANK[JBPOINTER]] == ACCESS_KEY[JBANK[JBPOINTER+8'd1]]
		| PROTECTED_MEMORY == 1'b0))
		JBANK[JBPOINTER]<=JBANK[JBPOINTER+8'd1];
	else if ( (state == JSR3) & RDY & (MMU_STATE == 2'b11) 
		& (ACCESS_KEY[JBANK[JBPOINTER]] == 8'd0 
		| ACCESS_KEY[JBANK[JBPOINTER]] == ACCESS_KEY[JBANK[JBPOINTER+8'd1]]
		| PROTECTED_MEMORY == 1'b0))
		JBPOINTER<=JBPOINTER+8'd1;
	else if ( (state == RTS1) & RDY  & (JBPOINTER>8'd0) & (MMU_STATE == 2'b11)
		& (ACCESS_KEY[JBANK[JBPOINTER-8'd1]] == 8'd0 
		| ACCESS_KEY[JBANK[JBPOINTER]] == ACCESS_KEY[JBANK[JBPOINTER-8'd1]]
		| PROTECTED_MEMORY == 1'b0))
		JBPOINTER<=JBPOINTER-8'd1;
	else if ( MMU_STATE < 2'b11)
		BANK_RANGE <= 4'bzzzz;
	else if (BANK_REG_ACT == 1'b1)
		case (BANK_REG_ADDR[8:0])
			9'd0: FBANK <= BANK_REG_DATA;
			9'd1: ZBANK[JBANK[JBPOINTER]] <= BANK_REG_DATA;
			9'd2: ZBANK[(JBPOINTER>8'd0)?(JBANK[JBPOINTER-8'd1]):JBANK[JBPOINTER]] <= BANK_REG_DATA;
			9'd3: IBANK <= BANK_REG_DATA;
			9'd4: NBANK <= BANK_REG_DATA;
			9'd5: JBPOINTER <= BANK_REG_DATA;
			9'b1xxxxxxxx: JBANK[AB[7:0]] <= BANK_REG_DATA;
		endcase


/*
 * register select logic. This determines which of the A, X, Y or
 * S registers will be accessed. 
 */

always @*  
    case( state )
        INDY1,
        INDX0,
        ZPX0,
        JMPIX0,
        ABSX0  : regsel = index_y ? SEL_Y : SEL_X;


        DECODE : regsel = dst_reg; 

        BRK0,
        BRK3,
        JSR0,
        JSR2,
        PULL0,
        PULL1,
        PUSH1,
        RTI0,
        RTI3,
        RTS0,
        RTS2   : regsel = SEL_S;
        
        default: regsel = src_reg; 
    endcase

/*
 * ALU
 */

ALU ALU( .clk(clk),
         .op(alu_op),
         .right(alu_shift_right),
         .AI(AI),
         .BI(BI),
         .CI(CI),
         .BCD(adc_bcd & (state == FETCH)),
         .CO(CO),
         .OUT(ADD),
         .V(AV),
         .Z(AZ),
         .N(AN),
         .HC(HC),
         .RDY(RDY) );

/*
 * Select current ALU operation
 */

always @*
    case( state )
        READ:   alu_op = op;

		ZB1:	alu_op = (bc_bit==8'd0)?((bs_bit==8'hff)?1'bx:OP_OR):OP_AND;

        ZB3,
		BRA1:   alu_op = backwards ? OP_SUB : OP_ADD; 

        FETCH,
        REG :   alu_op = op; 

        DECODE,
        ABS1:   alu_op = 1'bx;

        PUSH1,
        BRK0,
        BRK1,
        BRK2,
        JSR0,
        JSR1:   alu_op = OP_SUB;

     default:   alu_op = OP_ADD;
    endcase

/*
 * Determine shift right signal to ALU
 */

always @*
    if( state == FETCH || state == REG || state == READ )
        alu_shift_right = shift_right;
    else
        alu_shift_right = 0;

/*
 * Sign extend branch offset.  
 */

always @(posedge clk)
    if( RDY )
        backwards <= DIMUX[7];

/* 
 * ALU A Input MUX 
 */

always @*
    case( state )
        JSR1,
        RTS1,
        RTI1,
        RTI2,
        BRK1,
        BRK2,
        INDX1:  AI = ADD;

        REG,
        ZPX0,
        INDX0,
        JMPIX0,
        ABSX0,
        RTI0,
        RTS0,
        JSR0,
        JSR2,
        BRK0,
        PULL0,
        INDY1,
        PUSH0,
        PUSH1:  AI = regfile;

        ZB1,
		ZB2,
		BRA0,
        READ:   AI = DIMUX;

		ZB3,
        BRA1:   AI = ABH;       // don't use PCH in case we're 

        FETCH:  AI = load_only ? 0 : regfile;

        DECODE,
        ABS1:   AI = 8'hxx;     // don't care

        default:  AI = 0;
    endcase


/*
 * ALU B Input mux
 */

always @*
    case( state )
		 ZB3,
         BRA1,
         RTS1,
         RTI0,
         RTI1,
         RTI2,
         INDX1,
         REG,
         JSR0,
         JSR1,
         JSR2,
         BRK0,
         BRK1,
         BRK2,
         PUSH0, 
         PUSH1,
         PULL0,
         RTS0:  BI = 8'h00;

         ZB1,
		 READ:  BI = txb_ins ? (trb_ins ? ~regfile : regfile) : 
				(set_bit == 8'd0)? ((clear_bit == 8'hff)? 
				((bc_bit == 8'd0)? (bs_bit == 8'hff ? 8'h00 : bs_bit)
				:bc_bit) : clear_bit) : set_bit;

         ZB2,
		 BRA0:  BI = PCL;

         DECODE,
         ABS1:  BI = 8'hxx;

         default:       BI = DIMUX;
    endcase

/*
 * ALU CI (carry in) mux
 */

always @*
    case( state )
        INDY2,
		ZB3,
        BRA1,
        JMPIX1,
        ABSX1:  CI = CO;

        DECODE,
        ABS1:   CI = 1'bx;

        READ,
        REG:    CI = rotate ? C :
                     shift ? 0 : inc;

        FETCH:  CI = rotate  ? C : 
                     compare ? 1 : 
                     (shift | load_only) ? 0 : C;

        PULL0,
        RTI0,
        RTI1,
        RTI2,
        RTS0,
        RTS1,
        INDY0,
        INDX1:  CI = 1; 

        default:        CI = 0;
    endcase

/*
 * Processor Status Register update
 *
 */

/*
 * Update C flag when doing ADC/SBC, shift/rotate, compare
 */
always @(posedge clk )
    if( shift && state == WRITE ) 
        C <= CO;
    else if( state == RTI2 )
        C <= DIMUX[0];
    else if( ~write_back && state == DECODE ) begin
        if( adc_sbc | shift | compare )
            C <= CO;
        else if( plp )
            C <= ADD[0];
        else begin
            if( sec ) C <= 1;
            if( clc ) C <= 0;
        end
    end

/*
 * Special Z flag got TRB/TSB
 */ 
always @(posedge clk) 
    AZ2 <= ~|(AI & regfile);

/*
 * Update Z, N flags when writing A, X, Y, Memory, or when doing compare
 */

always @(posedge clk) 
    if( state == WRITE)
        Z <= txb_ins ? AZ2 : AZ1;
    else if( state == RTI2 )
        Z <= DIMUX[1];
    else if( state == DECODE ) begin
        if( plp )
            Z <= ADD[1];
        else if( (load_reg & (regsel != SEL_S)) | compare | bit_ins )
            Z <= AZ1;
    end

always @(posedge clk)
    if( state == WRITE && ~txb_ins)
        N <= AN1;
    else if( state == RTI2 )
        N <= DIMUX[7];
    else if( state == DECODE ) begin
        if( plp )
            N <= ADD[7];
        else if( (load_reg & (regsel != SEL_S)) | compare )
            N <= AN1;
    end else if( state == FETCH && bit_ins_nv )
        N <= DIMUX[7];

/*
 * Update I flag
 */

always @(posedge clk)
    if( state == BRK3 )
        I <= 1;
    else if( state == RTI2 )
        I <= DIMUX[2];
    else if( state == REG ) begin
        if( sei ) I <= 1;
        if( cli ) I <= 0;
    end else if( state == DECODE )
        if( plp ) I <= ADD[2];

/*
 * Update D flag
 */
always @(posedge clk ) 
    if( state == RTI2 )
        D <= DIMUX[3];
    else if( state == DECODE ) begin
        if( sed ) D <= 1;
        if( cld ) D <= 0;
        if( plp ) D <= ADD[3];
    end

/*
 * Update V flag
 */
always @(posedge clk )
    if( state == RTI2 ) 
        V <= DIMUX[6];
    else if( state == DECODE ) begin
        if( adc_sbc ) V <= AV;
        if( clv )     V <= 0;
        if( plp )     V <= ADD[6];
    end else if( state == FETCH && bit_ins_nv )
        V <= DIMUX[6];

/*
 * Instruction decoder
 */

/*
 * IR register/mux. Hold previous DI value in IRHOLD in PULL0 and PUSH0
 * states. In these states, the IR has been prefetched, and there is no
 * time to read the IR again before the next decode.
 */

//reg RDY1 = 1;

//always @(posedge clk )
//    RDY1 <= RDY;

//always @(posedge clk )
//    if( ~RDY && RDY1 )
//        DIHOLD <= DI;

always @(posedge clk )
	if( reset == 1'b1 )
        IRHOLD_valid <= 0;
    else if( RDY ) begin
        if( state == PULL0 || state == PUSH0 ) begin
            IRHOLD <= DIMUX;
            IRHOLD_valid <= 1;
        end else if( state == DECODE )
            IRHOLD_valid <= 0;
    end

assign IR = (IRQ & ~I) | NMI_edge ? 8'h00 :
                     IRHOLD_valid ? IRHOLD : DIMUX;
					 

//assign DIMUX = ~RDY1 ? DIHOLD : DI;
   
//assign DIMUX = DI;   

/*
 * Input from memory unless BANK_REG_MEM=ON in which case the input comes from registers
 */
assign DIMUX = (BANK_REG_MEM == 1'b0) ? DI: 
		(AB[15:8] == 8'h01) ? JBANK[AB[7:0]]:
		(AB[15:0] == 16'h0000) ? FBANK:
		(AB[15:0] == 16'h0001) ? ZBANK[JBANK[JBPOINTER]]:
		(AB[15:0] == 16'h0002) ? ZBANK[JBANK[(JBPOINTER > 8'd0) ? JBPOINTER-8'd1:8'd0]]:
		(AB[15:0] == 16'h0003) ? IBANK:
		(AB[15:0] == 16'h0004) ? NBANK:
		(AB[15:0] == 16'h0005) ? JBPOINTER: DI;

/*
 * Microcode state machine
 */
always @(posedge clk or posedge reset)
	if( reset == 1'b1 )
        state <= BRK0;
    else if( RDY ) case( state )
        DECODE  : 
            casex ( IR )
                // TODO Review for simplifications as in verilog the first matching case has priority
                8'b0000_0000:   state <= BRK0;
                8'b0010_0000:   state <= JSR0;
                8'b0010_1100:   state <= ABS0;  // BIT abs
                8'b1001_1100:   state <= ABS0;  // STZ abs
                8'b000x_1100:   state <= ABS0;  // TSB/TRB
                8'b0100_0000:   state <= RTI0;  // 
                8'b0100_1100:   state <= JMP0;
                8'b0110_0000:   state <= RTS0;
                8'b0110_1100:   state <= JMPI0;
                8'b0111_1100:   state <= JMPIX0;
				8'b1101_1011:   state <= STP;	// STP instruction (stop execution)
				8'b1100_1011:   state <= WAI0;	// WAI instruction (wait for interrupt)
`ifdef IMPLEMENT_MMU
                8'bxxx0_0010:   state <= FETCH; // (NOP2: 2 column, 4 column handled correctly below)
                8'bx1x1_1100:   state <= ABS0;  // (NOP3: C column)
`endif
`ifdef IMPLEMENT_NOPS             
                8'bxxxx_x011:   state <= REG;   // (NOP1: 3/B column)
                8'bxxx0_0010:   state <= FETCH; // (NOP2: 2 column, 4 column handled correctly below)
                8'bx1x1_1100:   state <= ABS0;  // (NOP3: C column)
`endif
                8'b0x00_1000:   state <= PUSH0;
                8'b0x10_1000:   state <= PULL0;
                8'b0xx1_1000:   state <= REG;   // CLC, SEC, CLI, SEI 
                8'b11x0_00x0:   state <= FETCH; // IMM
                8'b1x10_00x0:   state <= FETCH; // IMM
                8'b1xx0_1100:   state <= ABS0;  // X/Y abs
                8'b1xxx_1000:   state <= REG;   // DEY, TYA, ... 
                8'bxxx0_0001:   state <= INDX0;
                8'bxxx1_0010:   state <= IND0;  // (ZP) odd 2 column
                8'b000x_0100:   state <= ZP0;   // TSB/TRB
				8'bxxxx_0111:   state <= ZP0;   // RMB/SMB (7 column)
                8'bxxxx_1111:   state <= ZB0;   // BBR/BBS (F column)
                8'bxxx0_01xx:   state <= ZP0;   // (also $44 MMF/MMJ)
                8'bxxx0_1001:   state <= FETCH; // IMM
                8'bxxx0_1101:   state <= ABS0;  // even D column
                8'bxxx0_1110:   state <= ABS0;  // even E column
                8'bxxx1_0000:   state <= BRA0;  // odd 0 column (Branches)
                8'b1000_0000:   state <= BRA0;  // BRA
                8'bxxx1_0001:   state <= INDY0; // odd 1 column
                8'bxxx1_01xx:   state <= ZPX0;  // odd 4,5,6,7 columns (also $54 MMJ/MMF)
                8'bxxx1_1001:   state <= ABSX0; // odd 9 column
                8'bx011_1100:   state <= ABSX0; // C column BIT (3C), LDY (BC)
                8'bxxx1_11x1:   state <= ABSX0; // odd D, F columns
                8'bxxx1_111x:   state <= ABSX0; // odd E, F columns
                8'bx101_1010:   state <= PUSH0; // PHX/PHY
                8'bx111_1010:   state <= PULL0; // PLX/PLY
                8'bx0xx_1010:   state <= REG;   // <shift> A, TXA, ...  NOP
                8'bxxx0_1010:   state <= REG;   // <shift> A, TXA, ...  NOP
            endcase

        ZP0     : state <= write_back ? READ : FETCH;

        ZPX0    : state <= ZPX1;
        ZPX1    : state <= write_back ? READ : FETCH;

        ABS0    : state <= write_back ? READ : ABS1;
        ABS1    : state <= FETCH;

        ABSX0   : state <= ABSX1;
        ABSX1   : state <= (CO | store | write_back) ? ABSX2 : FETCH;
        ABSX2   : state <= write_back ? READ : FETCH;

        JMPIX0  : state <= JMPIX1;
        JMPIX1  : state <= CO ? JMPIX2 : JMP0;
        JMPIX2  : state <= JMP0;

        IND0    : state <= INDX1;

        INDX0   : state <= INDX1;
        INDX1   : state <= INDX2;
        INDX2   : state <= INDX3;
        INDX3   : state <= FETCH;

        INDY0   : state <= INDY1;
        INDY1   : state <= INDY2;
        INDY2   : state <= (CO | store) ? INDY3 : FETCH;
        INDY3   : state <= FETCH;

        READ    : state <= WRITE;
        WRITE   : state <= FETCH;
        FETCH   : state <= DECODE;

        REG     : state <= DECODE;
        
        PUSH0   : state <= PUSH1;
        PUSH1   : state <= DECODE;

        PULL0   : state <= PULL1;
        PULL1   : state <= PULL2; 
        PULL2   : state <= DECODE;

        JSR0    : state <= JSR1;
        JSR1    : state <= JSR2;
        JSR2    : state <= JSR3;
        JSR3    : state <= FETCH; 

        RTI0    : state <= RTI1;
        RTI1    : state <= RTI2;
        RTI2    : state <= RTI3;
        RTI3    : state <= RTI4;
        RTI4    : state <= DECODE;

        RTS0    : state <= RTS1;
        RTS1    : state <= RTS2;
        RTS2    : state <= RTS3;
        RTS3    : state <= FETCH;

        WAI0    : state <= WAI1;
        WAI1    : state <= WAI2;
        WAI2    : state <= FETCH;

		ZB0     : state <= ZB1;
		ZB1     : state <= ZB2;
		ZB2     : state <= ((cond_save[4:4]==0)&&(ADD==8'd0)) ? ZB3 : 
							((cond_save[4:4]==1)&&(ADD==8'hff))? ZB3: ZB4;
		ZB3     : state <= (CO ^ backwards) ? ZB4 : FETCH;
		ZB4     : state <= FETCH;
		
        BRA0    : state <= cond_true ? BRA1 : DECODE;
        BRA1    : state <= (CO ^ backwards) ? BRA2 : DECODE;
        BRA2    : state <= DECODE;

        JMP0    : state <= JMP1;
        JMP1    : state <= DECODE; 

        JMPI0   : state <= JMPI1;
        JMPI1   : state <= JMP0;

        BRK0    : state <= BRK1;
        BRK1    : state <= BRK2;
        BRK2    : state <= BRK3;
        BRK3    : state <= JMP0;

        MMJ0    : state <= DECODE;
        MMJ1    : state <= FETCH;
        MMF0    : state <= DECODE; 
        MMF1    : state <= FETCH; 


    endcase

/*
 * Additional control signals
 */

always @(posedge clk)
	if( reset == 1'b1 )
         res <= 1;
     else if( state == DECODE )
         res <= 0;

always @(posedge clk)
     if( state == DECODE && RDY )
        casex( IR )             // DMB: Checked for 65C02 NOP collisions
                8'b0xx1_0010,   // ORA, AND, EOR, ADC (zp)
                8'b1x11_0010,   // LDA, SBC (zp)
                8'b0xxx_1010,   // ASLA, INCA, ROLA, DECA, LSRA, PHY, RORA, PLY
                8'b0xxx_xx01,   // ORA, AND, EOR, ADC
                8'b100x_10x0,   // DEY, TYA, TXA, TXS
                8'b1010_xxx0,   // LDA/LDX/LDY 
                8'b1011_1010,   // TSX
                8'b1011_x1x0,   // LDX/LDY
                8'b1100_1010,   // DEX
                8'b11x1_1010,   // PHX, PLX
                8'b1x1x_xx01,   // LDA, SBC
                8'bxxx0_1000:   // PHP, PLP, PHA, PLA, DEY, TAY, INY, INX
                                load_reg <= 1;

                default:        load_reg <= 0;
        endcase

always @(posedge clk)
     if( state == DECODE && RDY )
        casex( IR )
                8'b1110_1000,   // INX
                8'b1100_1010,   // DEX
                8'b1111_1010,   // PLX
                8'b1010_0010,   // LDX imm
                8'b101x_x110,   // LDX
                8'b101x_1x10:   // LDX, TAX, TSX
                                dst_reg <= SEL_X;

                8'b0x00_1000,   // PHP, PHA
                8'bx101_1010,   // PHX, PHY
                8'b1001_1010:   // TXS
                                dst_reg <= SEL_S;

                8'b1x00_1000,   // DEY, DEX
                8'b0111_1010,   // PLY
                8'b101x_x100,   // LDY
                8'b1010_x000:   // LDY #imm, TAY
                                dst_reg <= SEL_Y;

                default:        dst_reg <= SEL_A;
        endcase

always @(posedge clk)
     if( state == DECODE && RDY )
        casex( IR )
                8'b1011_1010:   // TSX 
                                src_reg <= SEL_S; 

                8'b100x_x110,   // STX
                8'b100x_1x10,   // TXA, TXS
                8'b1110_xx00,   // INX, CPX
                8'b1101_1010,   // PHX
                8'b1100_1010:   // DEX
                                src_reg <= SEL_X; 

                8'b100x_x100,   // STY
                8'b1001_1000,   // TYA
                8'b1100_xx00,   // CPY
                8'b0101_1010,   // PHY
                8'b1x00_1000:   // DEY, INY
                                src_reg <= SEL_Y;

                default:        src_reg <= SEL_A;
        endcase

always @(posedge clk) 
     if( state == DECODE && RDY )
        casex( IR )
                8'bxxx1_0001,   // INDY
                8'b10x1_0110,   // LDX zp,Y / STX zp,Y
                8'b1011_1110,   // LDX abs,Y
                8'bxxxx_1001:   // abs, Y
                                index_y <= 1;

                default:        index_y <= 0;
        endcase


always @(posedge clk)
     if( state == DECODE && RDY )
        casex( IR )             // DMB: Checked for 65C02 NOP collisions
                8'b1001_0010,   // STA (zp)
                8'b100x_x1x0,   // STX, STY, STZ abs, STZ abs,x
                8'b011x_0100,   // STZ zp, STZ zp,x
                8'b100x_xx01:   // STA
                                store <= 1;

                default:        store <= 0;

        endcase

always @(posedge clk )
     if( state == DECODE && RDY )
        casex( IR )             // DMB: Checked for 65C02 NOP collisions
                8'b0xxx_x110,   // ASL, ROL, LSR, ROR
                8'b000x_x100,   // TSB/TRB
                8'b11xx_x110,   // DEC/INC
                8'bxxxx_0111:   // RMB/SMB (column 7)				
                                write_back <= 1;

                default:        write_back <= 0;
        endcase


always @(posedge clk )
     if( state == DECODE && RDY )
        casex( IR )
                8'b101x_xxxx:   // LDA, LDX, LDY
                                load_only <= 1;
                default:        load_only <= 0;
        endcase

always @(posedge clk )
     if( state == DECODE && RDY )
        casex( IR )
                8'b0001_1010,   // INCA
                8'b111x_x110,   // INC 
                8'b11x0_1000:   // INX, INY
                                inc <= 1;

                default:        inc <= 0;
        endcase

always @(posedge clk )
     if( state == DECODE && RDY )
        casex( IR )
                8'b0000_0111:   clear_bit <= 8'b11111110;		// RMB0
                8'b0001_0111:   clear_bit <= 8'b11111101;		// RMB0
                8'b0010_0111:   clear_bit <= 8'b11111011;		// RMB0
                8'b0011_0111:   clear_bit <= 8'b11110111;		// RMB0
                8'b0100_0111:   clear_bit <= 8'b11101111;		// RMB0
                8'b0101_0111:   clear_bit <= 8'b11011111;		// RMB0
                8'b0110_0111:   clear_bit <= 8'b10111111;		// RMB0
                8'b0111_0111:   clear_bit <= 8'b01111111;		// RMB0

                default:        clear_bit <= 8'hff;
        endcase

always @(posedge clk )
     if( state == DECODE && RDY )
        casex( IR )
                8'b1000_0111:   set_bit <= 8'b00000001;		// SMB0
                8'b1001_0111:   set_bit <= 8'b00000010;		// SMB0
                8'b1010_0111:   set_bit <= 8'b00000100;		// SMB0
                8'b1011_0111:   set_bit <= 8'b00001000;		// SMB0
                8'b1100_0111:   set_bit <= 8'b00010000;		// SMB0
                8'b1101_0111:   set_bit <= 8'b00100000;		// SMB0
                8'b1110_0111:   set_bit <= 8'b01000000;		// SMB0
                8'b1111_0111:   set_bit <= 8'b10000000;		// SMB0

                default:        set_bit <= 8'd0;
        endcase

always @(posedge clk )
     if( state == DECODE && RDY )
        casex( IR )
                8'b1000_1111:   bs_bit <= 8'b11111110;		// BBS0
                8'b1001_1111:   bs_bit <= 8'b11111101;		// BBS0
                8'b1010_1111:   bs_bit <= 8'b11111011;		// BBS0
                8'b1011_1111:   bs_bit <= 8'b11110111;		// BBS0
                8'b1100_1111:   bs_bit <= 8'b11101111;		// BBS0
                8'b1101_1111:   bs_bit <= 8'b11011111;		// BBS0
                8'b1110_1111:   bs_bit <= 8'b10111111;		// BBS0
                8'b1111_1111:   bs_bit <= 8'b01111111;		// BBS0

                default:        bs_bit <= 8'hff;
        endcase

always @(posedge clk )
     if( state == DECODE && RDY )
        casex( IR )
                8'b0000_1111:   bc_bit <= 8'b00000001;		// BBR0
                8'b0001_1111:   bc_bit <= 8'b00000010;		// BBR0
                8'b0010_1111:   bc_bit <= 8'b00000100;		// BBR0
                8'b0011_1111:   bc_bit <= 8'b00001000;		// BBR0
                8'b0100_1111:   bc_bit <= 8'b00010000;		// BBR0
                8'b0101_1111:   bc_bit <= 8'b00100000;		// BBR0
                8'b0110_1111:   bc_bit <= 8'b01000000;		// BBR0
                8'b0111_1111:   bc_bit <= 8'b10000000;		// BBR0

                default:        bc_bit <= 8'd0;
        endcase

always @(posedge clk )
     if( (state == DECODE || state == BRK0) && RDY )
        casex( IR )
                8'bx111_0010,   // SBC (zp), ADC (zp)
                8'bx11x_xx01:   // SBC, ADC
                                adc_sbc <= 1;

                default:        adc_sbc <= 0;
        endcase

always @(posedge clk )
     if( (state == DECODE || state == BRK0) && RDY )
        casex( IR )
                8'b0111_0010,   // ADC (zp)
                8'b011x_xx01:   // ADC
                                adc_bcd <= D;

                default:        adc_bcd <= 0;
        endcase

always @(posedge clk )
     if( state == DECODE && RDY )
        casex( IR )
                8'b0xxx_x110,   // ASL, ROL, LSR, ROR (abs, absx, zpg, zpgx)
                8'b0xx0_1010:   // ASL, ROL, LSR, ROR (acc)
                                shift <= 1;

                default:        shift <= 0;
        endcase

always @(posedge clk )
     if( state == DECODE && RDY )
        casex( IR )
                8'b1101_0010,   // CMP (zp)
                8'b11x0_0x00,   // CPX, CPY (imm/zp)
                8'b11x0_1100,   // CPX, CPY (abs)
                8'b110x_xx01:   // CMP 
                                compare <= 1;

                default:        compare <= 0;
        endcase

always @(posedge clk )
     if( state == DECODE && RDY )
        casex( IR )
                8'b01xx_x110,   // ROR, LSR
                8'b01xx_1x10:   // ROR, LSR
                                shift_right <= 1;

                default:        shift_right <= 0; 
        endcase

always @(posedge clk )
     if( state == DECODE && RDY )
        casex( IR )
                8'b0x10_1010,   // ROL A, ROR A
                8'b0x1x_x110:   // ROR, ROL 
                                rotate <= 1;

                default:        rotate <= 0; 
        endcase

always @(posedge clk )
     if( state == DECODE && RDY )
        casex( IR )
                8'b0000_x100,   // TSB
                8'b1xxx_0111:   // SMB (column 7)				
                //8'b1xxx_1111:   // BBS (column F)	
                                op <= OP_OR;

                8'b0001_x100:   // TRB
                                op <= OP_AND;

                8'b00xx_x110,   // ROL, ASL
                8'b00x0_1010:   // ROL, ASL
                                op <= OP_ROL;

                8'b1000_1001,   // BIT imm
                8'b001x_x100,   // BIT zp/abs/zpx/absx
                8'b0xxx_0111:   // RMB (column 7)				
                //8'b0xxx_1111:   // BBR (column F)	
								op <= OP_AND;

                8'b01xx_x110,   // ROR, LSR
                8'b01xx_1x10:   // ROR, LSR
                                op <= OP_A;

                8'b11x1_0010,   // CMP, SBC (zp)
                8'b0011_1010,   // DEC A
                8'b1000_1000,   // DEY
                8'b1100_1010,   // DEX 
                8'b110x_x110,   // DEC 
                8'b11xx_xx01,   // CMP, SBC
                8'b11x0_0x00,   // CPX, CPY (imm, zpg)
                8'b11x0_1100:   op <= OP_SUB;

                8'b00x1_0010,   // ORA, AND (zp)
                8'b0x01_0010,   // ORA, EOR (zp)
                8'b010x_xx01,   // EOR
                8'b00xx_xx01:   // ORA, AND
                                op <= { 2'b11, IR[6:5] };
                
                default:        op <= OP_ADD; 
        endcase

always @(posedge clk )
     if( state == DECODE && RDY )
        casex( IR )
                8'b001x_x100:   // BIT zp/abs/zpx/absx (update N,V,Z)
                                {bit_ins, bit_ins_nv}  <= 2'b11;

                8'b1000_1001:   // BIT imm (update Z)
                                {bit_ins, bit_ins_nv}  <= 2'b10;

                default:        // not a BIT instruction
                                {bit_ins, bit_ins_nv}  <= 2'b00;
        endcase

always @(posedge clk )
     if( state == DECODE && RDY )
        casex( IR )
                8'b000x_x100:   // TRB/TSB
                                txb_ins <= 1;

                default:        txb_ins <= 0;
        endcase

always @(posedge clk )
     if( state == DECODE && RDY )
        casex( IR )
                8'b0001_x100:   // TRB
                                trb_ins <= 1;

                default:        trb_ins <= 0;
        endcase

always @(posedge clk )
     if( state == DECODE && RDY )
        casex( IR )
                8'b1001_11x0,   // STZ abs, STZ abs,x
                8'b011x_0100:   // STZ zp, STZ zp,x
                                store_zero <= 1;

                default:        store_zero <= 0;
        endcase

always @(posedge clk )
	 if( reset == 1'b1 )
        mmu_fetch <= 2'd0;
     else if( state == DECODE && RDY )
        casex( IR )
                8'b1111_1100:   // MMF
                                mmu_fetch <= 2'd2;		// Pick from $ADDR not data
                8'b0100_0100:   // MMF 
                                mmu_fetch <= 2'd1;		// Pick from $ZP data

                default:        mmu_fetch <= 2'd0;
        endcase

always @(posedge clk )
	 if ( reset == 1'b1 )
        mmu_jump <= 2'd0;
	 else if( state == DECODE && RDY )
        casex( IR )
                8'b1101_1100:   // MMJ
                                mmu_jump <= 2'd2;		// Pick from $ADDR not data				
                8'b0101_0100:   // MMJ 
                                mmu_jump <= 2'd1;		// Pick from $ZP data

                default:        mmu_jump <= 2'd0;
        endcase

always @(posedge clk )
	 if ( reset == 1'b1 )
		 mmu_special <= 3'b000;
     else if( state == DECODE && RDY )
        casex( IR )
                8'b1000_0010:   // MMS
                                mmu_special <= 3'b001;
                8'b1100_0010:   // MMZ 
                                mmu_special <= 3'b010;
                8'b1110_0010:   // MMB 
                                mmu_special <= 3'b011;
				8'b0101_1100:   // MMI 
                                mmu_special <= 3'b100;

                default:        mmu_special <= 3'b000;
        endcase



/*
 * special instructions
 */
always @(posedge clk )
     if( state == DECODE && RDY ) begin
        php <= (IR == 8'h08);
        clc <= (IR == 8'h18);
        plp <= (IR == 8'h28);
        sec <= (IR == 8'h38);
        cli <= (IR == 8'h58);
        sei <= (IR == 8'h78);
        clv <= (IR == 8'hb8);
        cld <= (IR == 8'hd8);
        sed <= (IR == 8'hf8);
        brk <= (IR == 8'h00);
     end

always @(posedge clk)
	if ( reset == 1'b1 )
		cond_save <= 5'b00000;
    else if( RDY )
		begin
			cond_code <= {IR[7:4]};	// Leftmost bits of instruction, BBR/BBS
			if (IR[3:0]==4'hf)
				cond_save <= {IR[7:4],((IR[3:0]==4'hf)?1'b1:1'b0)};	// Leftmost bits of instruction, BBR/BBS=1
		end


always @*
    casex( cond_code )
            4'b0001: cond_true = ~N;
            4'b0011: cond_true = N;
            4'b0101: cond_true = ~V;
            4'b0111: cond_true = V;
            4'b1001: cond_true = ~C;
            4'b1011: cond_true = C;
            4'b1101: cond_true = ~Z;
            4'b1111: cond_true = Z;
            default: cond_true = 1; // BRA is 80
    endcase

reg NMI_1 = 0;          // delayed NMI signal

always @(posedge clk)
    NMI_1 <= NMI;

always @(posedge clk )
    if( NMI_edge && state == BRK3 )
        NMI_edge <= 0;
    else if( NMI & ~NMI_1 )
        NMI_edge <= 1;

endmodule



